"""
This package implements the API to feed a powerpoint presentation as a template
to a source of data. The actual template contents can be templated out with
your favourite template engine.
"""

from .core import *  # noqa
